import csv
import datetime
import zoneinfo
from decimal import Decimal
from io import BytesIO
from unittest import mock

from lxml import etree

from nem12_tools.generators import nem12
from nem12_tools.parsers.nmid import Meter, MeterPoint, Register


def test_nem12():
    m = MeterPoint(
        nmi="4102335210",
        role_mdp="ACTIVMDP",
        role_frmp="ENERGEX",
        meters=[
            Meter(
                serial_number="701226207",
                registers=[Register(register_id="E1", uom="KWH", suffix="E1")],
            )
        ],
    )
    notification = nem12.generate_nem12(m)
    xml_file = BytesIO()
    notification.tree.write(xml_file, pretty_print=True, xml_declaration=True, encoding="utf-8")

    root = etree.fromstring(xml_file.getvalue())
    assert root.findtext("./Header/From") == "ACTIVMDP"
    assert root.findtext("./Header/To") == "ENERGEX"
    assert root.findtext("./Header/TransactionGroup") == "MTRD"
    assert root.findtext("./Header/Priority") == "Medium"
    assert root.findtext("./Header/Market") == "NEM"

    csv_data = root.findtext(".//CSVIntervalData")
    assert csv_data is not None
    reader = csv.reader(csv_data.splitlines())
    section_100 = next(reader)
    assert section_100 == ["100", "NEM12", mock.ANY, "ACTIVMDP", "ENERGEX"]
    section_200 = next(reader)
    assert section_200 == [
        "200",
        "4102335210",
        "E1",
        "E1",
        "E1",
        "",
        "701226207",
        "KWH",
        "5",
        "",
    ]
    section_300 = next(reader)
    assert section_300[0] == "300"
    assert len(section_300) == 295
    assert section_300[-5:] == ["A", "", "", mock.ANY, mock.ANY]
    section_900 = next(reader)
    assert section_900 == ["900"]


def test_multiple_dates():
    m = MeterPoint(
        nmi="4102335210",
        role_mdp="ACTIVMDP",
        role_frmp="ENERGEX",
        meters=[
            Meter(
                serial_number="701226207",
                registers=[Register(register_id="E1", uom="KWH", suffix="E1")],
            )
        ],
    )
    notification = nem12.generate_nem12(
        m, start=datetime.date(2024, 1, 1), end=datetime.date(2024, 1, 2)
    )
    xml_file = BytesIO()
    notification.tree.write(xml_file, pretty_print=True, xml_declaration=True, encoding="utf-8")

    root = etree.fromstring(xml_file.getvalue())
    csv_data = root.findtext(".//CSVIntervalData")
    assert csv_data is not None
    reader = csv.reader(csv_data.splitlines())
    assert next(reader)[0] == "100"
    assert next(reader)[0] == "200"
    # each read date is nested under the 200 section rather than alternating
    assert next(reader)[0] == "300"
    assert next(reader)[0] == "300"
    assert next(reader)[0] == "900"


def test_interval_15():
    m = MeterPoint(
        nmi="4102335210",
        role_mdp="ACTIVMDP",
        role_frmp="ENERGEX",
        meters=[
            Meter(
                serial_number="701226207",
                registers=[Register(register_id="E1", uom="KWH", suffix="E1")],
            )
        ],
    )
    notification = nem12.generate_nem12(m, interval=nem12.IntervalLength.FIFTEEN_MINUTES)
    xml_file = BytesIO()
    notification.tree.write(xml_file, pretty_print=True, xml_declaration=True, encoding="utf-8")

    root = etree.fromstring(xml_file.getvalue())
    csv_data = root.findtext(".//CSVIntervalData")
    assert csv_data is not None
    reader = csv.reader(csv_data.splitlines())
    assert next(reader)[0] == "100"
    assert next(reader)[0] == "200"
    row_300 = next(reader)
    assert row_300[0] == "300"
    assert len(row_300) == 103


class TestHeader:
    def test_emits_row(self):
        now = datetime.datetime.now()
        now_formatted = now.strftime("%Y%m%d%H%M")
        header = nem12.Header(generation_time=now, from_participant="A", to_participant="B")
        assert header.as_row() == ("100", "NEM12", now_formatted, "A", "B")


class TestNmiDetails:
    def test_emits_row(self):
        nmi_details = nem12.NmiDetails(
            nmi="4102335210",
            nmi_configuration="E1",
            register_id="E1",
            register_suffix="E1",
            meter_serial_number="701226207",
            uom="KWH",
            interval_length=nem12.IntervalLength.FIVE_MINUTES,
        )
        assert nmi_details.as_row() == (
            "200",
            "4102335210",
            "E1",
            "E1",
            "E1",
            "",
            "701226207",
            "KWH",
            "5",
            "",
        )


class TestIntervalData:
    def test_emits_row(self):
        now = datetime.datetime(2024, 9, 3, 12, 34, 56)
        now_formatted = "20240903123456"
        interval_data = nem12.IntervalData(
            read_date=now.date(),
            read_values=(Decimal("1.0000"), Decimal("0.0500")),
            quality_method=nem12.QualityMethod.ACTUAL,
            last_updated=now,
            msats_load_time=now,
        )
        assert interval_data.as_row() == (
            "300",
            "20240903",
            "1.0000",
            "0.0500",
            "A",
            "",
            "",
            now_formatted,
            now_formatted,
        )


class TestTerminator:
    def test_emits_row(self):
        terminator = nem12.Terminator()
        assert terminator.as_row() == ("900",)


class TestProduceNem12Data:
    def test_interval_5_minutes(self):
        now = datetime.datetime(2024, 9, 3, 12, 34, 56, tzinfo=zoneinfo.ZoneInfo("Etc/GMT-10"))
        now_date = now.date()
        m = MeterPoint(
            nmi="4102335210",
            role_mdp="ACTIVMDP",
            role_frmp="ENERGEX",
            meters=[
                Meter(
                    serial_number="701226207",
                    registers=[
                        Register(register_id="E1", uom="KWH", suffix="E1"),
                        Register(register_id="B1", uom="KWH", suffix="B1"),
                    ],
                )
            ],
        )
        data = nem12.produce_nem12_data(
            m, now_date, now_date, nem12.IntervalLength.FIVE_MINUTES, now
        )
        assert data.header.indicator == "100"

        (r1, [*r1_interval]), (r2, [*r2_interval]) = data.read_data
        assert r1.indicator == "200"
        assert r1.nmi == "4102335210"
        assert r1.register_id == "E1"
        assert len(r1_interval) == 1
        for interval in r1_interval:
            assert interval.read_date == now_date
            assert len(interval.read_values) == 288
            assert interval.quality_method == nem12.QualityMethod.ACTUAL
            assert interval.last_updated == now
            assert interval.msats_load_time == now
            value: Decimal
            for value in interval.read_values:
                assert len(str(value).split(".")[0]) == 1
                assert len(str(value).split(".")[1]) == 4

        assert r2.indicator == "200"
        assert r2.nmi == "4102335210"
        assert r2.register_id == "B1"
        assert len(r2_interval) == 1
        for interval in r2_interval:
            assert interval.read_date == now_date
            assert len(interval.read_values) == 288
            assert interval.quality_method == nem12.QualityMethod.ACTUAL
            assert interval.last_updated == now
            assert interval.msats_load_time == now
            value: Decimal
            for value in interval.read_values:
                assert len(str(value).split(".")[0]) == 1
                assert len(str(value).split(".")[1]) == 4

    def test_interval_15_minutes(self):
        now = datetime.datetime(2024, 9, 3, 12, 34, 56, tzinfo=zoneinfo.ZoneInfo("Etc/GMT-10"))
        now_date = now.date()
        m = MeterPoint(
            nmi="4102335210",
            role_mdp="ACTIVMDP",
            role_frmp="ENERGEX",
            meters=[
                Meter(
                    serial_number="701226207",
                    registers=[
                        Register(register_id="E1", uom="KWH", suffix="E1"),
                        Register(register_id="B1", uom="KWH", suffix="B1"),
                    ],
                )
            ],
        )
        data = nem12.produce_nem12_data(
            m, now_date, now_date, nem12.IntervalLength.FIFTEEN_MINUTES, now
        )
        assert data.header.indicator == "100"

        (r1, [*r1_interval]), (r2, [*r2_interval]) = data.read_data
        assert r1.indicator == "200"
        assert r1.nmi == "4102335210"
        assert r1.register_id == "E1"
        assert len(r1_interval) == 1
        for interval in r1_interval:
            assert interval.read_date == now_date
            assert len(interval.read_values) == 96
            assert interval.quality_method == nem12.QualityMethod.ACTUAL
            assert interval.last_updated == now
            assert interval.msats_load_time == now
            value: Decimal
            for value in interval.read_values:
                assert len(str(value).split(".")[0]) == 1
                assert len(str(value).split(".")[1]) == 4

        assert r2.indicator == "200"
        assert r2.nmi == "4102335210"
        assert r2.register_id == "B1"
        assert len(r2_interval) == 1
        for interval in r2_interval:
            assert interval.read_date == now_date
            assert len(interval.read_values) == 96
            assert interval.quality_method == nem12.QualityMethod.ACTUAL
            assert interval.last_updated == now
            assert interval.msats_load_time == now
            value: Decimal
            for value in interval.read_values:
                assert len(str(value).split(".")[0]) == 1
                assert len(str(value).split(".")[1]) == 4

    def test_interval_30_minutes(self):
        now = datetime.datetime(2024, 9, 3, 12, 34, 56, tzinfo=zoneinfo.ZoneInfo("Etc/GMT-10"))
        now_date = now.date()
        m = MeterPoint(
            nmi="4102335210",
            role_mdp="ACTIVMDP",
            role_frmp="ENERGEX",
            meters=[
                Meter(
                    serial_number="701226207",
                    registers=[
                        Register(register_id="E1", uom="KWH", suffix="E1"),
                        Register(register_id="B1", uom="KWH", suffix="B1"),
                    ],
                )
            ],
        )
        data = nem12.produce_nem12_data(
            m, now_date, now_date, nem12.IntervalLength.THIRTY_MINUTES, now
        )
        assert data.header.indicator == "100"

        (r1, [*r1_interval]), (r2, [*r2_interval]) = data.read_data
        assert r1.indicator == "200"
        assert r1.nmi == "4102335210"
        assert r1.register_id == "E1"
        assert len(r1_interval) == 1
        for interval in r1_interval:
            assert interval.read_date == now_date
            assert len(interval.read_values) == 48
            assert interval.quality_method == nem12.QualityMethod.ACTUAL
            assert interval.last_updated == now
            assert interval.msats_load_time == now
            value: Decimal
            for value in interval.read_values:
                assert len(str(value).split(".")[0]) == 1
                assert len(str(value).split(".")[1]) == 4

        assert r2.indicator == "200"
        assert r2.nmi == "4102335210"
        assert r2.register_id == "B1"
        assert len(r2_interval) == 1
        for interval in r2_interval:
            assert interval.read_date == now_date
            assert len(interval.read_values) == 48
            assert interval.quality_method == nem12.QualityMethod.ACTUAL
            assert interval.last_updated == now
            assert interval.msats_load_time == now
            value: Decimal
            for value in interval.read_values:
                assert len(str(value).split(".")[0]) == 1
                assert len(str(value).split(".")[1]) == 4



from xml.etree import ElementTree as ET

from lxml import etree


class MeterDataNotification:
    parser: etree.XMLParser
    root: etree._Element
    tree: etree._ElementTree

    def __init__(self):
        self.parser = etree.XMLParser(
            remove_blank_text=True
        )  # This will remove all ignorable whitespace and let the subsequent pretty-printing "start from scratch".
        self.root = etree.fromstring(self.xml_root(), self.parser)
        self.tree = etree.ElementTree(self.root)

    def xml_root(self):
        NS1 = "urn:aseXML:r43"
        NS2 = "http://www.w3.org/2001/XMLSchema-instance"
        ET.register_namespace("ase", NS1)
        ET.register_namespace("xsi", NS2)
        qname1 = ET.QName(NS1, "aseXML")  # Element QName
        qname2 = ET.QName(NS2, "schemaLocation")  # Attribute QName
        root = ET.Element(
            qname1.text,
            {
                qname2.text: "urn:aseXML:r43 http://www.nemmco.com.au/aseXML/schemas/r43/aseXML_r43.xsd"
            },
        )
        return ET.tostring(root)

    def header(
        self,
        from_text,
        to_text,
        message_id,
        message_date,
        transaction_group,
        priority,
        market,
    ):
        """
        <Header>
        <From>TXUN</From>
        <To>LOLOL</To>
        <MessageID>TXUN 3009a20200227120425987</MessageID>
        <MessageDate>2020-02-27T12:00:26+10:00</MessageDate>
        <TransactionGroup>SORD</TransactionGroup>
        <Priority>Low</Priority>
        <Market>VICGAS</Market>
        </Header>
        """
        self.header_parent = etree.SubElement(self.root, "Header")

        self.header_from = etree.SubElement(self.header_parent, "From")
        self.header_from.text = from_text

        self.header_to = etree.SubElement(self.header_parent, "To")
        self.header_to.text = to_text

        self.header_msg_id = etree.SubElement(self.header_parent, "messageID")
        self.header_msg_id.text = message_id

        self.header_msg_date = etree.SubElement(self.header_parent, "MessageDate")
        self.header_msg_date.text = message_date

        self.header_trans_grp = etree.SubElement(self.header_parent, "TransactionGroup")
        self.header_trans_grp.text = transaction_group

        self.header_priority = etree.SubElement(self.header_parent, "Priority")
        self.header_priority.text = priority

        self.header_market = etree.SubElement(self.header_parent, "Market")
        self.header_market.text = market

    def transactions(
        self,
        transaction_id,
        transaction_date,
        transaction_type,
        transaction_schema_version,
        csv_interval_data,
        participant_role,
    ):
        self.transactions_parent = etree.SubElement(self.root, "Transactions")
        self.transaction = etree.SubElement(
            self.transactions_parent,
            "Transaction",
            attrib={
                "transactionID": transaction_id,
                "transactionDate": transaction_date,
            },
        )
        self.transaction_type = etree.SubElement(
            self.transaction,
            transaction_type,
            attrib={"version": transaction_schema_version},
        )

        self.interval_data = etree.SubElement(self.transaction_type, "CSVIntervalData")
        self.interval_data.text = csv_interval_data

        self.trans_participant = etree.SubElement(
            self.transaction_type, "ParticipantRole"
        )

        self.trans_participant_role = etree.SubElement(self.trans_participant, "Role")
        self.trans_participant_role.text = participant_role

        """
      <Transactions>
        <Transaction transactionID="MTRD_MSG_NEM12_202112271020" transactionDate="2021-12-27T10:20:05.000+10:00">
          <MeterDataNotification version="r25">
            <CSVIntervalData>
            </CSVIntervalData>
              <ParticipantRole>
              </ParticipantRole>
          </MeterDataNotification>
        </Transaction>
      </Transactions>
      """

    def write_xml(self, output_filename: str):
        self.tree.write(
            output_filename,
            pretty_print=True,
            xml_declaration=True,
            encoding="utf-8",
        )



"""
Parse NMI Discovery details required for NEM12 file generation.
"""

import dataclasses

from lxml import etree


@dataclasses.dataclass()
class Register:
    register_id: str
    uom: str
    suffix: str


@dataclasses.dataclass()
class Meter:
    serial_number: str

    registers: list[Register]


@dataclasses.dataclass()
class MeterPoint:
    """
    NMI details required for NEM12 file generation.
    """

    nmi: str
    role_mdp: str
    role_frmp: str

    meters: list[Meter]


def from_nmidiscovery(xml_doc: str) -> MeterPoint:
    """
    Parse NMI Discovery XML document and return a list of MeterPoint objects.
    """

    root = etree.fromstring(xml_doc)

    return MeterPoint(
        nmi=_get_nmi(root),
        role_mdp=_get_participant(root, "MDP"),
        role_frmp=_get_frmp(root),
        meters=_get_meters(root),
    )


def _get_nmi(root) -> str:
    nmi = root.findtext(".//NMI")
    if not nmi:
        raise ValueError("NMI not found in NMI Discovery XML.")
    return nmi


def _get_meters(root: etree._Element) -> list[Meter]:
    """
    Get the list of meters from the XML root element.
    """
    meters = []
    for meter in root.findall(".//Meter"):
        status = meter.findtext(".//Status")
        if status == "C":
            serial_number = meter.findtext(".//SerialNumber")
            if not serial_number:
                raise ValueError("Serial number not found in NMI Discovery XML.")

            registers = []
            for register in meter.findall(".//Register"):
                register_status = register.findtext(".//Status")
                if register_status == "C":
                    register_id = register.findtext(".//RegisterID")
                    uom = register.findtext(".//UnitOfMeasure")
                    suffix = register.findtext(".//Suffix")
                    if not register_id or not uom or not suffix:
                        raise ValueError(
                            "Register details not found in NMI Discovery XML."
                        )

                    registers.append(Register(register_id, uom, suffix))

            if registers:  # Check if there are any current registers
                meters.append(Meter(serial_number, registers))

    return meters


def _get_participant(root: etree._Element, role: str) -> str:
    """
    Get the participant from the XML root element.
    """
    for participants in root.findall(".//RoleAssignment"):
        if participants.findtext(".//Role") == role:
            if party := participants.findtext(".//Party"):
                return party

    raise ValueError(f"Participant with role {role} not found.")


def _get_frmp(root: etree._Element) -> str:
    """
    Get the FRMP role from the XML root element.

    We extract the To participant from the header.
    """
    frmp = root.findtext("./Header/To")
    if not frmp:
        raise ValueError("FRMP not found in NMI Discovery XML.")
    return frmp



import os

from nem12_tools.parsers import nmid


def test_nmidiscovery_parsed():
    here = os.path.dirname(os.path.realpath(__file__))
    xml = open(f"{here}/../examples/nmi-discovery.xml").read()
    parsed = nmid.from_nmidiscovery(xml)
    assert parsed.nmi == "4102335210"
    assert parsed.role_mdp == "ACTIVMDP"
    assert parsed.role_frmp == "ENERGEX"
    assert len(parsed.meters) == 1
    meter = parsed.meters[0]
    assert meter.serial_number == "701226207"
    assert len(meter.registers) == 1
    register = meter.registers[0]
    assert register.register_id == "E1"
    assert register.uom == "KWH"
    assert register.suffix == "E1"



import datetime
from typing import IO

import click

from nem12_tools.generators import nem12
from nem12_tools.parsers.nmid import from_nmidiscovery


@click.command()
@click.argument("nmi_discovery_file", type=click.File("r"))
@click.argument("output_file", type=click.File("wb"))
@click.option(
    "--from",
    "from_date",
    type=click.DateTime(),
    help="Date to generate reads from. Default: today",
)
@click.option(
    "--to",
    "to_date",
    type=click.DateTime(),
    help="Date to generate reads to. Default: today",
)
@click.option(
    "--frmp",
    type=str,
    help="The FRMP role to receive the NEM data. Default: the MDP role in the NMI Discovery file.",
)
@click.option(
    "--interval",
    type=click.Choice(["5", "15", "30"]),
    default="5",
    help="The interval length in minutes. Default: 5",
)
def generate(
    nmi_discovery_file: IO[str],
    output_file: IO[bytes],
    from_date: datetime.datetime | None,
    to_date: datetime.datetime | None,
    frmp: str | None,
    interval: str,
) -> None:
    if not from_date:
        from_date = datetime.datetime.now()
    if not to_date:
        to_date = datetime.datetime.now()
    meter_config = from_nmidiscovery(nmi_discovery_file.read())
    if frmp:
        meter_config.role_frmp = frmp
    interval_length = nem12.IntervalLength(int(interval))
    meter_data_transaction = nem12.generate_nem12(
        meter_config, from_date.date(), to_date.date(), interval_length
    )
    meter_data_transaction.tree.write(
        output_file, pretty_print=True, xml_declaration=True, encoding="utf-8"
    )
    click.echo("NEM12 file generated successfully")



import pathlib

from click.testing import CliRunner

from nem12_tools.cli import generate


def test_generate(tmp_path: pathlib.Path):
    nmi_discovery = pathlib.Path(__file__).parent.parent / "examples/nmi-discovery.xml"
    nmi_discovery.resolve()
    runner = CliRunner()
    output_file = tmp_path / "output.xml"
    result = runner.invoke(
        generate,
        [
            str(nmi_discovery),
            str(output_file),
            "--from",
            "2021-01-01",
            "--to",
            "2021-01-02",
            "--frmp",
            "TEST",
            "--interval",
            "5",
        ],
    )

    assert result.exit_code == 0, result.exception
    assert "NEM12 file generated successfully" in result.output
    assert output_file.read_text().startswith("<?xml")



import csv
import datetime
import enum
import io
import random
import zoneinfo
from abc import ABC, abstractmethod
from decimal import Decimal
from typing import Sequence

from pydantic import BaseModel, Field, field_serializer

from nem12_tools.parsers.nmid import MeterPoint

from . import notifications as mdmt


@enum.unique
class IntervalLength(enum.IntEnum):
    FIVE_MINUTES = 5
    FIFTEEN_MINUTES = 15
    THIRTY_MINUTES = 30

    def intervals(self) -> int:
        return (24 * 60) // self.value


@enum.unique
class QualityMethod(enum.StrEnum):
    ACTUAL = "A"
    SUBSTITUTE = "S"
    PERMANENT_SUBSTITUTE = "F"


class RowProducer(ABC):
    @abstractmethod
    def as_row(self) -> tuple[str, ...]: ...


class Header(RowProducer, BaseModel):
    indicator: str = "100"
    version: str = "NEM12"
    generation_time: datetime.datetime = Field(default_factory=datetime.datetime.now)
    from_participant: str
    to_participant: str

    @field_serializer("generation_time")
    def serialize_generation_time(self, generation_time: datetime.datetime) -> str:
        return generation_time.strftime("%Y%m%d%H%M")

    def as_row(self) -> tuple[str, ...]:
        data = self.model_dump()
        return (
            data["indicator"],
            data["version"],
            data["generation_time"],
            data["from_participant"],
            data["to_participant"],
        )


class NmiDetails(RowProducer, BaseModel):
    indicator: str = "200"
    nmi: str
    nmi_configuration: str
    register_id: str
    register_suffix: str
    mdm_data_stream: str = ""
    meter_serial_number: str
    uom: str
    interval_length: IntervalLength
    next_scheduled_read_date: None = None

    @field_serializer("next_scheduled_read_date")
    def serialize_next_scheduled_read_date(self, next_scheduled_read_date: None) -> str:
        return ""

    @field_serializer("interval_length")
    def serialize_interval_length(self, interval_length: IntervalLength) -> str:
        return str(interval_length.value)

    def as_row(self) -> tuple[str, ...]:
        data = self.model_dump()
        return (
            data["indicator"],
            data["nmi"],
            data["nmi_configuration"],
            data["register_id"],
            data["register_suffix"],
            data["mdm_data_stream"],
            data["meter_serial_number"],
            data["uom"],
            data["interval_length"],
            data["next_scheduled_read_date"],
        )


class IntervalData(RowProducer, BaseModel):
    indicator: str = "300"
    read_date: datetime.date
    read_values: tuple[Decimal, ...]
    quality_method: QualityMethod
    reason_code: str = ""
    reason_description: str = ""
    last_updated: datetime.datetime = Field(default_factory=datetime.datetime.now)
    msats_load_time: datetime.datetime = Field(default_factory=datetime.datetime.now)

    @field_serializer("read_date")
    def serialize_read_date(self, read_date: datetime.date) -> str:
        return read_date.strftime("%Y%m%d")

    @field_serializer("read_values")
    def serialize_read_values(self, read_values: list[Decimal]) -> tuple[str, ...]:
        return tuple(str(read) for read in read_values)

    @field_serializer("quality_method")
    def serialize_quality_method(self, quality_method: QualityMethod) -> str:
        return quality_method.value

    @field_serializer("last_updated")
    def serialize_last_updated(self, last_updated: datetime.datetime) -> str:
        return last_updated.strftime("%Y%m%d%H%M%S")

    @field_serializer("msats_load_time")
    def serialize_msats_load_time(self, msats_load_time: datetime.datetime) -> str:
        return msats_load_time.strftime("%Y%m%d%H%M%S")

    def as_row(self) -> tuple[str, ...]:
        data = self.model_dump()
        return (
            data["indicator"],
            data["read_date"],
            *data["read_values"],
            data["quality_method"],
            data["reason_code"],
            data["reason_description"],
            data["last_updated"],
            data["msats_load_time"],
        )


class Terminator(RowProducer, BaseModel):
    indicator: str = "900"

    def as_row(self) -> tuple[str, ...]:
        return (self.indicator,)


class Nem12Data(BaseModel):
    header: Header
    read_data: Sequence[tuple[NmiDetails, Sequence[IntervalData]]]
    terminator: Terminator


def generate_nem12(
    meter_point: MeterPoint,
    start: datetime.date = datetime.date.today(),
    end: datetime.date = datetime.date.today(),
    interval: IntervalLength = IntervalLength.FIVE_MINUTES,
) -> mdmt.MeterDataNotification:
    if start > end:
        raise ValueError("Start date must be before end date")

    now_tz = datetime.datetime.now(tz=zoneinfo.ZoneInfo("Etc/GMT-10"))
    nem_12_data = produce_nem12_data(meter_point, start, end, interval, now_tz)

    transactions = io.StringIO(newline="")
    writer = csv.writer(transactions, delimiter=",", lineterminator="\n")
    writer.writerow(nem_12_data.header.as_row())
    for nmi_details, interval_data in nem_12_data.read_data:
        writer.writerow(nmi_details.as_row())
        for data in interval_data:
            writer.writerow(data.as_row())
    writer.writerow(nem_12_data.terminator.as_row())

    meter_data_file = _create_meterdata_notification(meter_point)
    meter_data_file.transactions(
        transaction_id=f"MTRD_MSG_NEM12_{now_tz.strftime('%Y%m%d%H%M%f')}",
        transaction_date=now_tz.isoformat(timespec="seconds"),
        transaction_type="MeterDataNotification",
        transaction_schema_version="r25",
        csv_interval_data=transactions.getvalue(),
        participant_role="FRMP",
    )
    return meter_data_file


def produce_nem12_data(
    meter_point: MeterPoint,
    start: datetime.date,
    end: datetime.date,
    interval: IntervalLength,
    generation_time: datetime.datetime,
) -> Nem12Data:
    header = Header(
        generation_time=generation_time,
        from_participant=meter_point.role_mdp,
        to_participant=meter_point.role_frmp,
    )

    nmi_config = "".join(reg.suffix for meter in meter_point.meters for reg in meter.registers)
    read_data = []
    for meter in meter_point.meters:
        for register in meter.registers:
            nmi_details = NmiDetails(
                nmi=meter_point.nmi,
                nmi_configuration=nmi_config,
                register_id=register.register_id,
                register_suffix=register.suffix,
                meter_serial_number=meter.serial_number,
                uom=register.uom,
                interval_length=interval,
            )
            current_date = start
            interval_data = []
            while current_date <= end:
                interval_data.append(
                    IntervalData(
                        read_date=current_date,
                        read_values=_generate_consumption_profile(interval.intervals()),
                        quality_method=QualityMethod.ACTUAL,
                        last_updated=generation_time,
                        msats_load_time=generation_time,
                    )
                )
                current_date += datetime.timedelta(days=1)
            read_data.append((nmi_details, interval_data))

    return Nem12Data(header=header, read_data=read_data, terminator=Terminator())


def _generate_consumption_profile(
    intervals: int, min_value: float = -0.6, max_value: float = 0.8
) -> tuple[Decimal, ...]:
    """
    Generate reads over a 24 hour period over the given number of intervals.

    By default, we bias the read values towards 0 with a negative lower bound that we then max to 0.
    """
    # Generate a consumption profile with a bell shaped curve, peaking at approximately 8pm
    values = sorted(
        # Bias the numbers towards the mode
        round(max(0, random.triangular(min_value, max_value, mode=0.6)), 4)
        for _ in range(intervals)
    )
    # The pivot is selected to get to approximately 8pm
    pivot = int(intervals // 1.2)
    early, late = values[:pivot], values[pivot:]
    # Low consumption in the morning, getting higher towards 8pm
    early.sort()
    # Peak at about 8pm, then decreasing towards midnight
    late.sort(reverse=True)
    padding = Decimal("0.0000")
    return tuple(Decimal(f"{read}").quantize(padding) for read in early + late)


def _create_meterdata_notification(
    meter_point: MeterPoint,
) -> mdmt.MeterDataNotification:
    now_tz = datetime.datetime.now(tz=zoneinfo.ZoneInfo("Etc/GMT-10"))
    meter_data_file = mdmt.MeterDataNotification()
    meter_data_file.header(
        from_text=meter_point.role_mdp,
        to_text=meter_point.role_frmp,
        message_id=f"MTRD_MSG_NEM12_{now_tz.strftime('%Y%m%d%H%M%f')}",
        message_date=now_tz.isoformat(timespec="seconds"),
        transaction_group="MTRD",
        priority="Medium",
        market="NEM",
    )
    return meter_data_file
